from os.path import abspath, dirname, basename

import pkg_resources

try:
    with open(f'{abspath(dirname(__file__))}/../VERSION') as file:
        b_aws_testing_framework_version = file.read()
        b_aws_testing_framework_version = ''.join(b_aws_testing_framework_version.split())
except Exception:
    try:
        package_name = basename(dirname(__file__))
        b_aws_testing_framework_version = pkg_resources.get_distribution(package_name).version
    except Exception:
        b_aws_testing_framework_version = 'unspecified'



import logging
import time
from typing import Optional

from botocore.exceptions import ClientError

from b_aws_testing_framework.credentials import Credentials

logger = logging.getLogger(__name__)


class StackWaiter:
    """
    Waiter class that waits for a cloud formation stack to stabilize.
    """

    def __init__(self, stack_name: str):
        self.__stack_name = stack_name
        self.__client = Credentials().boto_session.client('cloudformation')

    def wait(self, current_iteration: int = 0, max_iterations: int = 100, sleep_time: int = 5) -> None:
        """
        Waits for a given stack to stabilize. E.g. waits when CREATE_IN_PROGRESS turns into CREATE_COMPLETE.

        :param current_iteration: This is a recursive function. This parameter indicates current iteration.
        :param max_iterations: Specifies maximum amount of iterations allowed.
        :param sleep_time: Specifies how long to sleep between iterations.

        :return: No return.
        """
        if current_iteration == max_iterations:
            raise RecursionError()

        stack_status = self.__get_stack_status() or ''
        logger.info(f'Stack status: {stack_status}.')

        if '_IN_PROGRESS' in stack_status:
            time.sleep(sleep_time)
            self.wait(current_iteration + 1, max_iterations, sleep_time)

    def __get_stack_status(self) -> Optional[str]:
        """
        Gets given stack status e.g. CREATE_IN_PROGRESS.

        :return: Stack status.
        """
        try:
            stack = self.__client.describe_stacks(StackName=self.__stack_name)['Stacks'][0]
            return stack['StackStatus']
        except ClientError as ex:
            # Retrieve code and message from an error.
            code = ex.response['Error']['Code']
            message = ex.response['Error']['Message']

            # If stack does not exist, its fine, just return None.
            if code == 'ValidationError' and 'does not exist' in message.lower():
                return

            # Otherwise throw an exception.
            raise



import logging
import os
import sys

# Append python package root path so resources could be imported.
# Failing to add these few lines below will result in a "module not found" error.
ROOT_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(ROOT_PATH)

from b_aws_testing_framework.credentials import Credentials
from b_aws_testing_framework.tools.cf_testing.cf_tool_config import CfToolConfig
from b_aws_testing_framework.tools.cf_testing.testing_manager import TestingManager

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)

STACK_PATH = f'{os.path.dirname(os.path.abspath(__file__))}/test_stack.yaml'


def pytest_configure(*args, **kwargs):
    """
    Called after command line options have been parsed and all plugins and initial conftest files been loaded.
    """
    TestingManager(Credentials(), CfToolConfig(STACK_PATH)).prepare_infrastructure()


def pytest_unconfigure(*args, **kwargs):
    """
    Called before test process is exited.
    """
    TestingManager(Credentials(), CfToolConfig(STACK_PATH)).destroy_infrastructure()



import logging

from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

logger = logging.getLogger(__name__)


def test_outputs() -> None:
    """
    Tests that outputs were created.

    :return: No return.
    """
    assert TestingStack.get_output('TestKey1') == 'TestValue1'
    assert TestingStack.get_output('TestKey2') == 'TestValue2'
    assert TestingStack.get_output('TestKey3') == 'TestValue3'



import logging

from b_aws_testing_framework.credentials import Credentials
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

logger = logging.getLogger(__name__)


def test_stack_exists() -> None:
    """
    Tests that stack actually was created.

    :return: No return.
    """
    client = Credentials().boto_session.client('cloudformation')

    stacks = client.list_stacks(StackStatusFilter=['CREATE_COMPLETE'])['StackSummaries']
    stacks = [stack['StackName'] for stack in stacks]

    assert TestingStack.name() in stacks



import logging
import os
import sys
from os.path import abspath, dirname, join
from typing import Any, Optional, Callable

from b_continuous_subprocess.continuous_subprocess import ContinuousSubprocess

from b_aws_testing_framework.base_testing_manager import BaseTestingManager
from b_aws_testing_framework.credentials import Credentials
from b_aws_testing_framework.tools.cdk_testing.cdk_tool_config import CdkToolConfig

logger = logging.getLogger(__name__)


class TestingManager(BaseTestingManager):
    """
    Test manager class which prepares infrastructure for tests.
    After tests are finished, destroys the infrastructure.
    """

    def __init__(self, credentials: Credentials, config: CdkToolConfig):
        super().__init__(credentials)

        self.__config = config

        # Ensure that project root is accessible for cdk process.
        if self.__config.project_root_path:
            sys.path.append(self.__config.project_root_path)

        # Additional process environment context.
        __additional_env = {
            **(self.credentials.environ or {}),
            **os.environ.copy(),
            **{'PYTHONPATH': ':'.join(sys.path)}
        }

        # Update the config environment with additional process environment context.
        for key, value in __additional_env.items():
            if config.deployment_process_environment.get(key) is None:
                config.deployment_process_environment[key] = value

    def prepare_infrastructure(self, custom_deploy_action: Optional[Callable[[CdkToolConfig], Any]] = None) -> None:
        """
        Prepares infrastructure to run tests.
        Firstly, the infrastructure is boot-strapped.
        Secondly, the infrastructure is destroyed (if any leftovers exist).
        Thirdly, the infrastructure is created.

        :param custom_deploy_action: Read more about this parameter in parent class.

        :return: No return.
        """
        logger.info('Setting global prefix (not overriding an existing one)...')
        self.set_global_prefix(override=False)

        logger.info('Bootstrapping the infrastructure...')
        self.__bootstrap_infrastructure()

        if self.__config.destroy_before_preparing:
            logger.info('Destroying the infrastructure before creating a new one...')
            self.__destroy_infrastructure()

        if custom_deploy_action:
            custom_deploy_action(self.__config)
        else:
            self.__create_infrastructure()

    def destroy_infrastructure(self, custom_destroy_action: Optional[Callable[[CdkToolConfig], Any]] = None) -> None:
        """
        Destroys the infrastructure.

        :param custom_destroy_action: Read more about this parameter in parent class.

        :return: No return.
        """
        if custom_destroy_action:
            custom_destroy_action(self.__config)
        else:
            self.__destroy_infrastructure()

        self.delete_global_prefix()

    """
    Infrastructure functions.
    """

    def __bootstrap_infrastructure(self) -> None:
        sub = ContinuousSubprocess(TestingManager.__aws_cdk_bootstrap_command())
        output = sub.execute(path=self.__config.cdk_app_path, env=self.__config.deployment_process_environment)
        for line in output: logger.info(line.strip())

    def __create_infrastructure(self) -> None:
        sub = ContinuousSubprocess(TestingManager.__aws_cdk_deploy_command())
        output = sub.execute(path=self.__config.cdk_app_path, env=self.__config.deployment_process_environment)
        for line in output: logger.info(line.strip())

    def __destroy_infrastructure(self) -> None:
        sub = ContinuousSubprocess(TestingManager.__aws_cdk_destroy_command())
        output = sub.execute(path=self.__config.cdk_app_path, env=self.__config.deployment_process_environment)
        for line in output: logger.info(line.strip())

    """
    CDK Commands.
    """

    @classmethod
    def __aws_cdk_bootstrap_command(cls) -> str:
        return f'{cls.custom_cdk_command()} bootstrap'

    @classmethod
    def __aws_cdk_deploy_command(cls) -> str:
        return f'{cls.custom_cdk_command()} deploy --all --require-approval never'

    @classmethod
    def __aws_cdk_destroy_command(cls) -> str:
        return f'{cls.custom_cdk_command()} destroy --all --require-approval never --force'

    """
    Helpers.
    """

    @classmethod
    def custom_cdk_command(cls) -> str:
        return join(abspath(dirname(__file__)), 'custom_cdk.sh')



class CfToolConfig:
    """
    Configuration class for cloud formation tool.
    """

    def __init__(self, cf_template_path: str):
        self.__cf_template_path = cf_template_path

    @property
    def cf_template_path(self):
        return self.__cf_template_path
